from pynvml import *
import sys
import helper_functions as main_funcs
import parse_args
import time


def main():
    
    # Getting a configuration obj
    config = parse_args.parse_cmd_args(sys.argv)

    if config.action == 'help':
        main_funcs.print_help()
        return

    try:

        # Verify driver version
        try:
            # Call nvml init and shutdown as an exception (special case), because I need it to restart in a loop later 
            nvmlInit()
            main_funcs.check_driver_version(nvmlSystemGetDriverVersion())
            nvmlShutdown()

        except main_funcs.UnsupportedDriverVersion:
            print('WARNING: You are running an unsupported driver, you may have problems')

        while(True):
            try:
                # Start nvml (or again in case of errors)
                # Useful to keep the process running during driver updates
                nvmlInit()
                
                match config.action:

                    # Information query
                    case 'list':
                        main_funcs.list_gpus()

                    case 'get-power-limit-info':
                        main_funcs.print_power_limit_info(config)

                    case 'get-thresholds-info':
                        main_funcs.print_thresholds_info(config)

                    case 'fan-info':
                        main_funcs.print_fan_info(config)

                    case 'fan-policy':
                        main_funcs.fan_policy(config)

                    case 'fan-policy-info':
                        main_funcs.print_fan_policy_info(config)

                    # Enable everything
                    case 'control':
                        main_funcs.control_all(config)
                    
                # If everything works fine, we don't need to retry
                break
                
            except Exception as error:
                print(f"Logging error: {error}")

                if config.retry == True:
                    # Clear all handles that nvml is holding, so errors won't persist across runs
                    nvmlShutdown()
                    print(f"Retrying in {config.retry_interval_s} seconds\n")
                    time.sleep(config.retry_interval_s)                

                else:
                    raise error
    
    # One should call shutdown with or without erros, this is why I am using finally
    finally:
        print('Calling nvml shutdown and terminating the program')
        nvmlShutdown()

if __name__ == '__main__':
    main()


import helper_functions
from helper_functions import error_print

class InvalidAction(Exception):
    pass

class InvalidOption(Exception):
    pass

class InvalidNumberSpeedPairParams(Exception):
    pass

class InvalidFanSpeed(Exception):
    pass

class InsufficientArgs(Exception):
    pass

class InvalidTimeParameter(Exception):
    pass

class InvalidConfig(Exception):
    pass


class Configuration:

    def __init__(self):
        # This only supports one target gpu, use a process for each GPU (erros become isolated to each other) 
        self.target_gpu = ""
        self.gpu_name = ""
        self.gpu_uuid = ""
        self.action = ""
        self.temp_speed_pair = []
        self.curve_type = "fixed" # Currently for internal usage only (I want to later add calculation for lines and curves fuctions)
        self.default_speed = 50 # Percentage
        self.time_interval = 1.0 # In seconds
        self.retry_interval_s = 2.0 # In seconds
        self.dry_run = False
        self.fan_policy = ''
        self.single_use = False
        self.acoustic_temp_limit = 0 # The user must set the value
        self.power_limit = 0 # The user must set the value
        self.verbose = False # Omit log messages by default
        self.retry = False # Let the service manager restart the process instead, only retry when asked explicitly

class TempSpeedPair:

    def __init__(self, temperature, speed):

        self.temperature = temperature
        self.speed = speed # Percentage!

    # The sorting is based on the temperature
    def __lt__(self, other):
        if (self.temperature <= other.temperature):
            return True

        else:
            return False

    # Important for testing
    def __eq__(self, other):
        if (self.temperature == other.temperature and self.speed == other.speed):
            return True

        else:
            return False

# Some sane checks (in case the user makes a bad config by accident)
def validate_config(config):

    # At least one of the target setting must be configured
    if config.gpu_name == '' and config.gpu_uuid == '':
        error_print("You did not select a target GPU")
        raise InvalidConfig("No GPU was selected")

    # fan-policy needs a mode
    if config.action == 'fan-policy':
        if config.fan_policy == '':
            error_print("You did not select a fan policy: autmatic or manual")
            raise InvalidConfig("No fan policy was selected")

    # temp-control needs a power limit configuration
    if config.action == 'control':
        if len(config.temp_speed_pair) == 0 and config.power_limit == 0 and config.acoustic_temp_limit == 0:
            error_print("You did not select any setting, please use one")
            raise InvalidConfig("No setting was selected")

        # Print warnings just to let users know
        if len(config.temp_speed_pair) == 0:
            print(f'WARNING: There is no temperature-speed pairs configured')

        if config.power_limit == 0:
            print(f'WARNING: There is no power limit configured')
        
        if config.acoustic_temp_limit == 0:
            print(f'WARNING: There is no temperature limit configured')


def parse_cmd_args(args):

    configuration = Configuration()

    if len(args) == 1:
        helper_functions.print_help()
        error_print(f'You must pass more arguments')
        raise InsufficientArgs("No action was supplied")

    # You can always ignore the first argument, since it is the program itself
    # Get the second arg, which is the action
    action = args[1]

    # The action names are decoupled from the cmd interface, allowing for flexibility
    if (action == 'help'):
        configuration.action = 'help'
        return configuration # It should stop here, ignore all other args

    elif (action == 'list'):
        configuration.action = 'list'
        return configuration # It should stop here, ignore all other args

    elif (action == 'fan-info'):
        configuration.action = 'fan-info'

    elif (action == 'fan-policy'):
        configuration.action = 'fan-policy'

    elif (action == 'fan-policy-info'):
        configuration.action = 'fan-policy-info'

    elif (action == 'power-limit-info'):
        configuration.action = 'get-power-limit-info'

    elif (action == 'thresholds-info'):
        configuration.action = 'get-thresholds-info'

    elif (action == 'control'):
        configuration.action = 'control'

    else:
        helper_functions.print_help()
        print(f'Invalid action: {action}\n\n')
        raise InvalidAction("The action passed as argument is incorrect")


    # You can safely ignore the action here
    i = 2
    while(i < len(args)):

        arg = args[i]

        if (arg == '--name' or arg == '-n'):
            configuration.gpu_name = args[i+1]
            i += 1 # Skip the next iteration

        elif (arg == '--uuid' or arg == '-id'):
            configuration.gpu_uuid = args[i+1]
            i += 1 # Skip the next iteration

        elif (arg == '--speed-pair' or arg == '-sp'):

            # Think of as points in a graph (speed % x temp °C)
            speed_points = args[i+1].split(',')

            for speed_pair_str in speed_points:
    
                speed_pair = speed_pair_str.split(':')

                if (len(speed_pair) != 2):
                    error_print('Invalid number of speed pair parameters')
                    raise InvalidNumberSpeedPairParams("You can only set temperature and target speed at a time")

                temp = int(speed_pair[0])
                speed = int(speed_pair[1])

                if (speed > 100):
                    error_print(f'The fan speed only goes up to 100%. You chose {speed}')
                    raise InvalidFanSpeed(f'The fan speed only goes up to 100%. You chose {speed}')

                if (speed < 0):
                    error_print(f'The fan speed cannot be lower than 0%. You chose {speed}')
                    raise InvalidFanSpeed(f'The fan speed cannot be lower than 0%. You chose {speed}')

                configuration.temp_speed_pair.append( TempSpeedPair(temp, speed) )

            i += 1 # Skip the next iteration

        elif (arg == '--default-speed' or arg == '-ds'):
            configuration.default_speed = int(args[i+1])
            i += 1 # Skip the next iteration

        elif (arg == '--time-interval' or arg == '-ti'):
            configuration.time_interval = float(args[i+1])
            i += 1 # Skip the next iteration

            # Refuse to continue with negative time values
            if configuration.time_interval < 0:
                error_print("You cannot use negative time values for intervals")
                raise InvalidTimeParameter("Invalid time parameter")

        elif (arg == '--retry-interval' or arg == '-ri'):
            configuration.retry_interval_s = float(args[i+1])
            i += 1 # Skip the next iteration

            # Refuse to continue with negative time values
            if configuration.retry_interval_s < 0:
                error_print("You cannot use negative time values for intervals")
                raise InvalidTimeParameter("Invalid time parameter")

        elif (arg == '--dry-run' or arg == '-dr'):
            configuration.dry_run = True

        elif (arg == '--verbose' or arg == '-V'):
            configuration.verbose = True

        elif (arg == '--retry' or arg == '-rt'):
            configuration.retry = True

        # For the fan-policy action
        elif (arg == '--auto'):
            configuration.fan_policy = 'automatic'

        elif (arg == '--manual'):
            configuration.fan_policy = 'manual'

        elif (arg == '--single-use' or arg == '-su'):
            configuration.single_use = True

        elif (arg == '--acoustic-temp-limit' or arg == '-tl'):
            configuration.acoustic_temp_limit = int(args[i+1])
            i += 1 # Skip the next iteration

        elif (arg == '--power-limit' or arg == '-pl'):
            configuration.power_limit = int(args[i+1])
            i += 1 # Skip the next iteration

        else:
            helper_functions.print_help()
            error_print(f'Invalid option: {arg}\n\n')
            raise InvalidOption('The option given was invalid')

        # Change iteration
        i += 1

    # Organizing the array before sending the configuration
    configuration.temp_speed_pair.sort(reverse=True)

    validate_config(configuration)

    return configuration


import sys
import os

# If we run the script from main (like python -m or directly from script), the other scripts might not find the imports
# So we add the dir where __main__ resides as an import path
# This also helps to keep the rest of the code clean
sys.path.append(os.path.dirname( os.path.realpath(__file__) ))

import nvml_gpu_control

# Call from python -m
if __name__ == '__main__':
    nvml_gpu_control.main()

# Call from script entrypoint
def script_call():
    nvml_gpu_control.main()


import unittest
from unittest.mock import Mock
import sys
import ctypes
sys.path.append('./src/caioh_nvml_gpu_control/') # Necessary so the tested files can all find each other from the projects root
import parse_args
import helper_functions as main_funcs

# Test command: python.exe .\tests.py -b

class TestMethods(unittest.TestCase):

    def test_parse_args_inssuficient_args(self):
        with self.assertRaises(parse_args.InsufficientArgs):
            parse_args.parse_cmd_args(['.python_script'])

    def test_parse_args_help(self):
        config = parse_args.parse_cmd_args(['.python_script', 'help'])
        self.assertEqual( config.action, 'help')

    def test_parse_args_list(self):
        config = parse_args.parse_cmd_args(['.python_script', 'list'])
        self.assertEqual( config.action, 'list')

    def test_parse_args_action_control_all(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')

    def test_parse_args_fan_policy(self):
        config = parse_args.parse_cmd_args(['.python_script', 'fan-policy', '--name', 'RTX 4080', '--auto'])
        self.assertEqual( config.action, 'fan-policy')

    def test_parse_args_fan_policy_info(self):
        config = parse_args.parse_cmd_args(['.python_script', 'fan-policy-info', '--name', 'RTX 4080'])
        self.assertEqual( config.action, 'fan-policy-info')

    def test_parse_args_fan_info(self):
        config = parse_args.parse_cmd_args(['.python_script', 'fan-info', '--name', 'RTX 4080'])
        self.assertEqual( config.action, 'fan-info')

    def test_parse_args_get_power_limit_info(self):
        config = parse_args.parse_cmd_args(['.python_script', 'power-limit-info', '--name', 'RTX 4080'])
        self.assertEqual( config.action, 'get-power-limit-info')

    def test_parse_args_get_temp_thresholds_limit_info(self):
        config = parse_args.parse_cmd_args(['.python_script', 'thresholds-info', '--name', 'RTX 4080'])
        self.assertEqual( config.action, 'get-thresholds-info')

    def test_parse_args_invalid_action(self):
        with self.assertRaises(parse_args.InvalidAction):
            parse_args.parse_cmd_args(['.python_script', 'invalid-action'])

    def test_parse_args_action_ignore_rest(self):
        config = parse_args.parse_cmd_args(['.python_script', 'help', '--name', 'RTX 4080'])
        self.assertEqual( config.action, 'help')
        self.assertEqual( config.target_gpu, '')

        config = parse_args.parse_cmd_args(['.python_script', 'list', '--name', 'RTX 4080'])
        self.assertEqual( config.action, 'list')
        self.assertEqual( config.target_gpu, '')

    def test_parse_args_option_gpu_name(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.gpu_name, 'RTX 4080')

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.gpu_name, 'RTX 3080')

    def test_parse_args_option_gpu_uuid(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--uuid', 'GPU-00000000-0000-0000-0000-000000000000', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.gpu_uuid, 'GPU-00000000-0000-0000-0000-000000000000')

    def test_parse_args_option_power_limit(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '--power-limit', '100'])
        self.assertEqual( config.power_limit, 100)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-pl', '100'])
        self.assertEqual( config.power_limit, 100)

    def test_parse_args_option_acoustic_temp_limit(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '--acoustic-temp-limit', '50'])
        self.assertEqual( config.acoustic_temp_limit, 50)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-tl', '50'])
        self.assertEqual( config.acoustic_temp_limit, 50)

    def test_parse_args_option_single_use(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-pl', '50', '--single-use'])
        self.assertEqual( config.single_use, True)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-pl', '50', '-su'])
        self.assertEqual( config.single_use, True)

        # Default should be False
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-pl', '50'])
        self.assertEqual( config.single_use, False)

    def test_parse_args_option_default_speed(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--default-speed', '36', '-n', 'RTX 3080', '-pl', '250'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.default_speed, 36)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-ds', '27', '-n', 'RTX 3080', '-pl', '250'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.default_speed, 27)
    
        config = parse_args.parse_cmd_args(['.python_script', 'control', '-n', 'RTX 3080', '-pl', '250'])
        self.assertEqual( config.action, 'control')
        self.assertTrue( config.default_speed >= 30) # Fan speed must never default for a value lower than 30%, except for when user explicitly wants to

    def test_parse_args_option_time_interval(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--time-interval', '5', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.time_interval, 5.0)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-ti', '0.5', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.time_interval, 0.5)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-ti', '0', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.time_interval, 0)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertTrue( config.time_interval <= 1) # Default should never be higher than 1s, unless the user states so

    def test_parse_args_option_time_interval_Error_negative_time(self):

        with self.assertRaises(parse_args.InvalidTimeParameter):
            parse_args.parse_cmd_args(['.python_script', 'control', '--time-interval', '-5', '-n', 'RTX 3080'])

    def test_parse_args_option_retry_interval(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--retry-interval', '5', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.retry_interval_s, 5.0)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-ri', '0.5', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.retry_interval_s, 0.5)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-ri', '0', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertEqual( config.retry_interval_s, 0)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual( config.action, 'control')
        self.assertTrue( config.retry_interval_s <= 2) # Default should never be higher than 2s, unless the user states so

    def test_parse_args_option_retry_interval_Error_negative_time(self):

        with self.assertRaises(parse_args.InvalidTimeParameter):
            parse_args.parse_cmd_args(['.python_script', 'control', '--retry-interval', '-5', '-n', 'RTX 3080'])

    def test_parse_args_temp_speed_pair(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair', '0:0,10:30,20:50,35:75,40:100', '-n', 'RTX 3080'])

        expected_output = [
            parse_args.TempSpeedPair(40, 100),
            parse_args.TempSpeedPair(35, 75),
            parse_args.TempSpeedPair(20, 50),
            parse_args.TempSpeedPair(10, 30),
            parse_args.TempSpeedPair(0, 0),
        ]
        self.assertEqual(expected_output, config.temp_speed_pair)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-sp', '0:0,10:30,20:50,35:75,40:100', '-n', 'RTX 3080'])
        self.assertEqual(expected_output, config.temp_speed_pair)

    def test_parse_args_temp_speed_pair_sort(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair', '40:100,20:50,10:30,35:75', '-n', 'RTX 3080'])

        expected_output = [
            parse_args.TempSpeedPair(40, 100),
            parse_args.TempSpeedPair(35, 75),
            parse_args.TempSpeedPair(20, 50),
            parse_args.TempSpeedPair(10, 30),
        ]
        self.assertEqual(expected_output, config.temp_speed_pair)

    def test_parse_args_temp_speed_pair_empty_list(self):

        with self.assertRaises(IndexError):
            parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair'])

    def test_parse_args_temp_speed_pair_invalid_list(self):

        with self.assertRaises(parse_args.InvalidNumberSpeedPairParams):
            parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair', ''])

        with self.assertRaises(parse_args.InvalidNumberSpeedPairParams):
            parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair', '10-20'])

        with self.assertRaises(ValueError):
            parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair', '10-20:10-20'])
        
        with self.assertRaises(ValueError):
            parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair', '10-20:10-20,10-20:10-20'])

    def test_parse_args_temp_speed_pair_invalid_fan_speed(self):

        with self.assertRaises(parse_args.InvalidFanSpeed):
            parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair', '10:120'])

        with self.assertRaises(parse_args.InvalidFanSpeed):
            parse_args.parse_cmd_args(['.python_script', 'control', '--speed-pair', '10:-100'])

    def test_parse_args_dry_run(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--dry-run', '-n', 'RTX 3080', '-pl', '250'])
        self.assertEqual(config.dry_run, True)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-dr', '-n', 'RTX 3080', '-pl', '250'])
        self.assertEqual(config.dry_run, True)

        # Default value should always be False
        config = parse_args.parse_cmd_args(['.python_script', 'control', '-n', 'RTX 3080', '-pl', '250'])
        self.assertEqual(config.dry_run, False)

    def test_parse_args_fan_policy_auto(self):
        config = parse_args.parse_cmd_args(['.python_script', 'fan-policy', '--name', 'RTX 4080', '--auto'])
        self.assertEqual( config.fan_policy, 'automatic')

    def test_parse_args_fan_policy_manual(self):
        config = parse_args.parse_cmd_args(['.python_script', 'fan-policy', '--name', 'RTX 4080', '--manual'])
        self.assertEqual( config.fan_policy, 'manual')

    def test_parse_args_option_verbose(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control','-n', 'RTX 3080', '--verbose', '-sp', '0:50'])
        self.assertEqual(config.verbose, True)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-n', 'RTX 3080', '-V', '-sp', '0:50'])
        self.assertEqual(config.verbose, True)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '-n', 'RTX 3080', '-sp', '0:50'])
        self.assertEqual(config.verbose, False)

    def test_parse_args_option_retry(self):
        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-pl', '50', '--retry'])
        self.assertEqual(config.retry, True)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-pl', '50', '-rt'])
        self.assertEqual(config.retry, True)

        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '-pl', '50'])
        self.assertEqual(config.retry, False)

    def test_parse_args_invalid_option(self):

        with self.assertRaises(parse_args.InvalidOption):
            parse_args.parse_cmd_args(['.python_script', 'control', '--invalid-option', '10:120', '-n', 'RTX 3080'])

    def test_parse_args_real_cmd(self):

        config = parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080', '--speed-pair', '0:0,20:35,30:50,40:100', '--time-interval', '0.5'])

        self.assertEqual(config.action, 'control')
        self.assertEqual(config.gpu_name, 'RTX 4080')
        self.assertEqual(config.time_interval, 0.5)
        expected_output = [
            parse_args.TempSpeedPair(40, 100),
            parse_args.TempSpeedPair(30, 50),
            parse_args.TempSpeedPair(20, 35),
            parse_args.TempSpeedPair(0, 0),
        ]
        self.assertEqual(config.temp_speed_pair, expected_output)

    def test_parse_args_sane_checks_no_fan_policy(self):

        # No fan policy
        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'fan-policy', '-n', 'RTX 3080'])

    def test_parse_args_sane_checks_no_gpu(self):

        # No target gpu
        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'control'])

        # No target gpu
        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'control'])

        # No target gpu
        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'control'])

        # No target gpu
        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'power-limit-info'])

        # No target gpu
        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'thresholds-info'])

    def test_parse_args_sane_checks_no_power_limit(self):

        # No fan policy
        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080'])

    def test_parse_args_sane_checks_no_temp_acoustic_limit(self):

        # No fan policy
        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'control', '--name', 'RTX 4080'])


    def test_check_driver_version(self):

        # If the driver starts to return letters, a refactoring will be needed anyways
        # So I just want to verify that letters reaise erros
        with self.assertRaises(ValueError):
            main_funcs.check_driver_version('AAA')

        # Only cares for the major version
        with self.assertRaises(main_funcs.UnsupportedDriverVersion):
            main_funcs.check_driver_version('515')

        with self.assertRaises(main_funcs.UnsupportedDriverVersion):
            main_funcs.check_driver_version('515.20.20')

        with self.assertRaises(main_funcs.UnsupportedDriverVersion):
            main_funcs.check_driver_version('515.20.20.20')

        with self.assertRaises(main_funcs.UnsupportedDriverVersion):
            main_funcs.check_driver_version('515.20.20.20aaaaa')

    def test_fan_policy_info_msg(self):
        msg = main_funcs.fan_policy_info_msg(ctypes.c_uint(0).value)
        self.assertEqual('Automatic', msg)

        msg = main_funcs.fan_policy_info_msg(ctypes.c_uint(1).value)
        self.assertEqual('Manual', msg)

        msg = main_funcs.fan_policy_info_msg(ctypes.c_uint(100).value)
        self.assertEqual('Unknown', msg)

    def test_parse_args_empty_config(self):

        with self.assertRaises(parse_args.InvalidConfig):
            parse_args.parse_cmd_args(['.python_script', 'control','--name', 'RTX 4080'])


#    # GPU Functions - I will need to improve the tests later
#
#    def test_gpu_something(self):
#        # Mocking
#        import pynvml
#
#        pynvml.nvmlDeviceGetCount = Mock(return_value=1)
#        pynvml.nvmlDeviceGetHandleByIndex = Mock(return_value=0)
#        pynvml.nvmlDeviceGetName = Mock(return_value='RTX 3080')
#
#        # Main function
#        main_funcs.list_gpus()
#
#        # Fail
#        self.assertTrue(True)



if __name__ == '__main__':
    unittest.main()


import pynvml
import datetime
import time
import ctypes
import sys

output_separator = '==============================================='

class UnsupportedDriverVersion(Exception):
    pass

class GpuNotFound(Exception):
    pass

class TemperatureThresholds:
    def __init__(self, shutdown_t, slowdown_t, max_memory_t, gpu_max_t, min_acoustic_t, current_acoustic_t, max_acoustic_t):
        self.shutdown = shutdown_t
        self.slowdown = slowdown_t
        self.max_memory = max_memory_t
        self.gpu_max = gpu_max_t
        self.min_acoustic = min_acoustic_t
        self.current_acoustic = current_acoustic_t
        self.max_acoustic = max_acoustic_t

class PowerLimitConstraintsWatts:
    def __init__(self, min_pl, max_pl):
        self.min = min_pl
        self.max = max_pl

class FanSpeedConstraintsPercentage:
    def __init__(self, min_s, max_s):
        self.min = min_s
        self.max = max_s

def check_driver_version(driver_version_str):
    major = int(driver_version_str.split('.')[0])

    if major < 520:
        raise UnsupportedDriverVersion('Driver version is lower than 520')

def log_helper(msg):
    print(f'LOG[{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]: {msg}')

def error_print(msg):
    print(f'{msg}', file=sys.stderr)

def print_help():
    help_text = '''
python ./nvml_gpu_control.py <ACTION> <OPTIONS>

ACTIONS
    help
          Display help text

    list
          List all available GPUs connected to the system by printing its name and UUID

    control
         Allows the use of all controls in a single command/loop. Each setting is enabled by configuring its respective option: fan curve, power and temperature

    fan-info
          Shows information about fan speed

    fan-policy <--auto|--manual>
          Changes the fan control policy to automatic (vBIOS controlled) or manual. Note that when the fan speed is changed, the NVML library automatically changes this setting to manual. This setting is useful to change the GPU back to its original state

    fan-policy-info
          Shows information about the current fan policy

    power-limit-info
          Shows information about the power limit of the selected GPU

    thresholds-info
          Shows information about temperature thresholds in dregrees Celsius of the selected GPU.

OPTIONS

    --name OR -n <GPU_NAME>
          Select a target GPU by its name. Example: --name "NVIDIA GeForce RTX 4080". Note: UUID has preference over name

    --uuid OR -id <GPU_UUID>
          Select a target GPU by its Universally Unique IDentifier (UUID). Example: --uuid "GPU-00000000-0000-0000-0000-000000000000". Note: UUID has preference over name

    --time-interval OR -ti <TIME_SECONDS>
          Time period in seconds to wait before probing the GPU again. Works for all actions that run in a loop

    --retry-interval OR -ri <TIME_SECONDS>
          Time period in seconds to wait before trying to issue commands to the GPU again. Works for all actions that run in a loop

    --dry-run OR -dr
          Run the program, but don't change/set anything. Useful for testing the behavior of the program

    --speed-pair OR -sp <TEMP_CELSIUS:SPEED_PERCENTAGE,TEMP_CELSIUS:SPEED_PERCENTAGE...>
          A comma separated list of pairs of temperature in celsius and the fan speed in % (temp:speed) defining basic settings for a fan curve

    --default-speed OR -ds <FAN_SPEED_PERCENTAGE>
          Set a default speed for when there is no match for the fan curve settings

    --manual
          Sets the fan policy to manual

    --auto
          Sets the fan policy to automatic (vBIOS controlled)

    --power-limit OR -pl <POWER_LIMIT_WATTS>
          Sets the power limit of the GPU in watts

    --acoustic-temp-limit OR -tl <TEMPERATURE_CELSIUS>
          Sets the acoustic threshold in celsious (note that this is the same temperature limit used by GeForce Experience)

    --single-use OR -su
          Makes some actions work only once instead of in a loop

    --verbose OR -V
          When there are no settings changes, leg messages are omitted by default. This option enables them back (good for debugging)

    --retry OR -rt
          By default the program closes when it encounters an error, but this lets the program try to recover. It could be useful to let the service manaer restart the process

'''
    print(help_text)

def list_gpus():
    deviceCount = pynvml.nvmlDeviceGetCount()

    for i in range(deviceCount):
        handle = pynvml.nvmlDeviceGetHandleByIndex(i)
        print(f'Device {i} name : {pynvml.nvmlDeviceGetName(handle)} - UUID: {pynvml.nvmlDeviceGetUUID(handle)}')

def print_GPU_info(gpu_handle):
    log_helper(f"Driver Version : {pynvml.nvmlSystemGetDriverVersion()}")
    log_helper(f"NVML Version : {pynvml.nvmlSystemGetNVMLVersion ()}")
    log_helper(f'Device name : {pynvml.nvmlDeviceGetName(gpu_handle)}')
    log_helper(f'Device UUID : {pynvml.nvmlDeviceGetUUID(gpu_handle)}')
    log_helper(f'Device fan speed : {pynvml.nvmlDeviceGetFanSpeed(gpu_handle)}%')
    log_helper(f'Fan policy : {fan_policy_info_msg( get_fan_policy(gpu_handle) )}')
    log_helper(f"Fan controller count : {pynvml.nvmlDeviceGetNumFans(gpu_handle)}")
    log_helper(f'Current temperature : {pynvml.nvmlDeviceGetTemperatureV(gpu_handle, pynvml.NVML_TEMPERATURE_GPU)}°C')
    log_helper(f'Temperature limit : {get_temperarure_thresholds(gpu_handle).current_acoustic}°C')
    log_helper(f'Power limit : {get_current_power_limit_watts(gpu_handle)}W')
    log_helper(f'Enforced power limit : {get_enforced_power_limit_watts(gpu_handle)}W')


# Search for a GPU and return a handle

def get_GPU_handle(gpu_name, gpu_uuid):
    
    if gpu_uuid != '':
        return pynvml.nvmlDeviceGetHandleByUUID(gpu_uuid)

    else:
        return get_GPU_handle_by_name(gpu_name)


# This will NOT work if the user has more than 2 GPUs with the same name/model, use UUID for this case
def get_GPU_handle_by_name(gpu_name):
    deviceCount = pynvml.nvmlDeviceGetCount()

    for i in range(deviceCount):
        handle = pynvml.nvmlDeviceGetHandleByIndex(i)

        if pynvml.nvmlDeviceGetName(handle) == gpu_name:
            return handle

    print(f'It was not possible to locate the target device : {gpu_name}')
    raise GpuNotFound('It was not possible to locate the device')

def set_gpu_fan_speed(gpu_handle, speed_percentage, dry_run):

    # This is not really the number of fan, but the number of controllers
    fan_count = pynvml.nvmlDeviceGetNumFans(gpu_handle)

    for fan_idx in range(fan_count):
        fan_speed = pynvml.nvmlDeviceGetFanSpeed_v2(gpu_handle, fan_idx)

        # Setting the fan speed DANGEROUS! Use dry run for testing before actual changes
        if dry_run != True:
            pynvml.nvmlDeviceSetFanSpeed_v2(gpu_handle, fan_idx, speed_percentage)

def get_gpu_fan_speed_per_controller(gpu_handle):

    fan_speed_per_controller = []

    # This is not really the number of fan, but the number of controllers
    fan_count = pynvml.nvmlDeviceGetNumFans(gpu_handle)

    for fan_idx in range(fan_count):
        fan_speed_per_controller.append(pynvml.nvmlDeviceGetFanSpeed_v2(gpu_handle, fan_idx))

    return fan_speed_per_controller

# This function will be here for only future references as many drivers ignore such values
def get_gpu_fan_speed_constraints(gpu_handle):

    # It needs pointers to work
    fan_min = ctypes.c_uint(0)
    fan_max = ctypes.c_uint(0)

    # Note some drivers do not respect the minimum and may turn off the fan motor in a different speed
    # Some drivers turn off the fan motor at speeds as high as 47%
    pynvml.nvmlDeviceGetMinMaxFanSpeed(gpu_handle, ctypes.byref(fan_min), ctypes.byref(fan_max))

    return FanSpeedConstraintsPercentage(fan_min.value, fan_max.value)

def print_fan_info(configuration):

    gpu_handle = get_GPU_handle(configuration.gpu_name, configuration.gpu_uuid)

    current_temp = pynvml.nvmlDeviceGetTemperatureV(gpu_handle, pynvml.NVML_TEMPERATURE_GPU)
    current_speed = pynvml.nvmlDeviceGetFanSpeed(gpu_handle)
    fan_constraints = get_gpu_fan_speed_constraints(gpu_handle)

    print(f'{output_separator}')
    print(f'Current temp: {current_temp}°C')
    print(f'Current speed: {current_speed}%') # Minitor for fan fan speed changes and reajust! 

    # Get the fan speed per controller
    for idx, fan_speed_c in enumerate(get_gpu_fan_speed_per_controller(gpu_handle)):
        print(f'Fan controller speed {idx}: {fan_speed_c}%')

    print(f'Fan constraints: Min {fan_constraints.min}% - Max {fan_constraints.max}%')
    print(f'{output_separator}')

# Control GPU functions and monitor for changes (e.g. temperature)
def fan_control_subroutine(gpu_handle, configuration):

    current_temp = pynvml.nvmlDeviceGetTemperatureV(gpu_handle, pynvml.NVML_TEMPERATURE_GPU)
    current_speed = pynvml.nvmlDeviceGetFanSpeed(gpu_handle)

    # Assume at first that it did change
    setting_changed = True

    log_msg = []

    log_msg.append(f'Current temp: {current_temp}°C')
    log_msg.append(f'Current speed: {current_speed}%') # Monitor for fan fan speed changes and reajust! 

    # Get the fan speed per controller
    for idx, fan_speed_c in enumerate(get_gpu_fan_speed_per_controller(gpu_handle)):
        log_msg.append(f'Fan controller speed {idx}: {fan_speed_c}%')

    for pair in configuration.temp_speed_pair:

        # Remember that that list starts by the highest temp value and keeps lowering it
        if current_temp >= pair.temperature:

            # Only send commands to the GPU if necessary (if the current setting is different from the targeted one)
            if current_speed != pair.speed:
                set_gpu_fan_speed(gpu_handle, pair.speed, configuration.dry_run)
                log_msg.append(f'Setting GPU fan speed: {pair.speed}%')
                setting_changed = True # Setting it again for safety
            else:
                log_msg.append(f'Same as previous speed, nothing to do!')
                setting_changed = False

            # Only print log messages when necessary to avoid taking too much disk space
            if configuration.verbose == True or setting_changed == True:
                log_helper(f'{"\n" + "\n".join(log_msg) + "\n"}')

            # Match found and set, return now
            return

    # We didn't find a match, use the default speed
    set_gpu_fan_speed(gpu_handle, configuration.default_speed, configuration.dry_run)
    log_helper(f'Found no temperature match, using default fan speed: {configuration.default_speed}')

def fan_policy_info_msg(fan_policy: int):

    if pynvml.NVML_FAN_POLICY_TEMPERATURE_CONTINOUS_SW == fan_policy:
        return 'Automatic'

    elif pynvml.NVML_FAN_POLICY_MANUAL == fan_policy:
        return 'Manual'

    else:
        return 'Unknown'

def set_fan_policy(gpu_handle, policy, dry_run):

    # This is not really the number of fan, but the number of controllers
    fan_count = pynvml.nvmlDeviceGetNumFans(gpu_handle)

    for fan_idx in range(fan_count):

        # Setting the fan control policy can be DANGEROUS! Use dry run for testing before actual changes
        if dry_run != True:
            fan_speed = pynvml.nvmlDeviceSetFanControlPolicy(gpu_handle, fan_idx, policy)

            # Also set the default fan speed for extra safety (automatic only)
            if policy == pynvml.NVML_FAN_POLICY_TEMPERATURE_CONTINOUS_SW:
                pynvml.nvmlDeviceSetDefaultFanSpeed_v2(gpu_handle, fan_idx)

def get_fan_policy(gpu_handle):
    current_policy = ctypes.c_uint(0)

    # The library unfortunately still needs pointers, this is why I need to use ctypes
    pynvml.nvmlDeviceGetFanControlPolicy_v2(gpu_handle, 0, ctypes.byref(current_policy))

    return current_policy.value


def print_fan_policy_info(configuration):
    gpu_handle = get_GPU_handle(configuration.gpu_name, configuration.gpu_uuid)
    print(f'{output_separator}')
    print(f'Current fan policy is: {fan_policy_info_msg( get_fan_policy(gpu_handle) )}')
    print(f'{output_separator}')

def fan_policy(configuration):

    target_fan_policy = configuration.fan_policy
    gpu_handle = get_GPU_handle(configuration.gpu_name, configuration.gpu_uuid)

    # Get the current policy before setting anything
    print(f'Current fan policy is: {fan_policy_info_msg( get_fan_policy(gpu_handle) )}')

    if target_fan_policy == 'automatic':
        set_fan_policy(gpu_handle, pynvml.NVML_FAN_POLICY_TEMPERATURE_CONTINOUS_SW, configuration.dry_run)

    elif target_fan_policy == 'manual':
        set_fan_policy(gpu_handle, pynvml.NVML_FAN_POLICY_MANUAL, configuration.dry_run)

    print('New fan control policy set sucessfully!')

    # Get the new policy
    print(f'Current fan policy is: {fan_policy_info_msg( get_fan_policy(gpu_handle) )}\n')

# Power control

# nvmlDeviceGetPowerManagementMode was deprecated: https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g10365092adc37d7a17d261db8fe63fb6
# nvmlDeviceSetPowerManagementLimit_v2 also seems to be deprecated, docs are non-existent: https://docs.nvidia.com/deploy/nvml-api/nvml_8h.html#nvml_8h_1d10040f340986af6cda91e71629edb2b

def set_power_limit(gpu_handle, power_limit_watts, dry_run):

    # Setting the power limit can be DANGEROUS! Use dry run for testing before actual changes
    if dry_run != True:
        pynvml.nvmlDeviceSetPowerManagementLimit(gpu_handle, int(power_limit_watts * 1000))

# Current power limit defined by the user, but it might defer from the enforced one
def get_current_power_limit_watts(gpu_handle):
    return int(pynvml.nvmlDeviceGetPowerManagementLimit(gpu_handle) / 1000)

# This one takes the constraints into account
def get_enforced_power_limit_watts(gpu_handle):
    return int(pynvml.nvmlDeviceGetEnforcedPowerLimit(gpu_handle) / 1000)

def get_power_limit_constraints_watts(gpu_handle):
    constraints_array = pynvml.nvmlDeviceGetPowerManagementLimitConstraints(gpu_handle)
    min = int(constraints_array[0] / 1000)
    max = int(constraints_array[1] / 1000)

    return PowerLimitConstraintsWatts(min, max)

def print_power_limit_info(configuration):
    gpu_handle = get_GPU_handle(configuration.gpu_name, configuration.gpu_uuid)

    constraints = get_power_limit_constraints_watts(gpu_handle)
    current_pl = get_current_power_limit_watts(gpu_handle)
    current_enforced_pl = get_enforced_power_limit_watts(gpu_handle)

    print(f'{output_separator}')
    print(f'Power limit constraints\nMin: {constraints.min}W - Max: {constraints.max}W\n')
    print(f'Current power limit: {current_pl}W\n')
    print(f'Current enforced power limit: {current_enforced_pl}W\n')
    print(f'{output_separator}')

def power_control_subroutine(gpu_handle, target_power_limit, dry_run, verbose):
    power_limit_constraints_watts = get_power_limit_constraints_watts(gpu_handle)
    current_pl = get_current_power_limit_watts(gpu_handle)
    current_enforced_pl = get_enforced_power_limit_watts(gpu_handle)

    setting_changed = True
    log_msg = []

    log_msg.append(f'Current power limit: {current_pl}W')
    log_msg.append(f'Current enforced power limit: {current_enforced_pl}W')

    if target_power_limit < power_limit_constraints_watts.min or target_power_limit > power_limit_constraints_watts.max:
        log_msg.append(f'WARNING: trying to set power limit outside of the min({power_limit_constraints_watts.min}W) and max({power_limit_constraints_watts.max}W) range')

    if target_power_limit != current_pl or target_power_limit != current_enforced_pl:
        set_power_limit(gpu_handle, target_power_limit, dry_run)
        log_msg.append(f'Setting the power limit: {target_power_limit}W')
        setting_changed = True

    else:
        log_msg.append(f'Nothing to do, current and enforced power limit is the same as the target')
        setting_changed = False

    # Only print log messages when necessary to avoid taking too much disk space
    if verbose == True or setting_changed == True:
        log_helper(f'{"\n" + "\n".join(log_msg) + "\n"}')

# Temperature control

# nvmlDeviceGetTemperatureThreshold is deprecated for some thresholds, use nvmlDeviceGetFieldValues insted
# https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g271ba78911494f33fc079b204a929405
def get_temperarure_thresholds(gpu_handle):
    # Info from nvidia-settings: T.Limit temperature after which GPU may shut down for HW protection
    #shutdown_threshold = pynvml.nvmlDeviceGetFieldValues(gpu_handle, [pynvml.NVML_FI_DEV_TEMPERATURE_SHUTDOWN_TLIMIT])[0].value.siVal
    shutdown_threshold = 0

    # Info from nvidia-settings: T.Limit temperature after which GPU may begin HW slowdown
    #slowdown_threshold = pynvml.nvmlDeviceGetFieldValues(gpu_handle, [pynvml.NVML_FI_DEV_TEMPERATURE_SLOWDOWN_TLIMIT])[0].value.siVal
    slowdown_threshold = 0

    # Info from nvidia-settings: T.Limit temperature after which GPU may begin SW slowdown due to memory temperature
    #max_memory_threshold = pynvml.nvmlDeviceGetFieldValues(gpu_handle, [pynvml.NVML_FI_DEV_TEMPERATURE_MEM_MAX_TLIMIT])[0].value.siVal
    max_memory_threshold = 0

    # Info from nvidia-settings: T.Limit temperature after which GPU may be throttled below base clock
    #gpu_max_threshold = pynvml.nvmlDeviceGetFieldValues(gpu_handle, [pynvml.NVML_FI_DEV_TEMPERATURE_GPU_MAX_TLIMIT])[0].value.siVal
    gpu_max_threshold =0

    # The acoustic settings is the same used by GeForce Experience
    # Info from nvidia-settings: Current temperature that is set as acoustic threshold.
    current_acoustic_threshold = pynvml.nvmlDeviceGetTemperatureThreshold(gpu_handle, pynvml.NVML_TEMPERATURE_THRESHOLD_ACOUSTIC_CURR)
    
    # These thresholds still use the old function
    # Info from nvidia-settings: Minimum GPU Temperature that can be set as acoustic threshold
    min_acoustic_threshold = pynvml.nvmlDeviceGetTemperatureThreshold(gpu_handle, pynvml.NVML_TEMPERATURE_THRESHOLD_ACOUSTIC_MIN)

    # Info from nvidia-settings: Maximum GPU temperature that can be set as acoustic threshold.
    max_acoustic_threshold = pynvml.nvmlDeviceGetTemperatureThreshold(gpu_handle, pynvml.NVML_TEMPERATURE_THRESHOLD_ACOUSTIC_MAX)

    return TemperatureThresholds(shutdown_threshold, slowdown_threshold, max_memory_threshold, gpu_max_threshold, min_acoustic_threshold, current_acoustic_threshold, max_acoustic_threshold)

# Valid values for threshold_type
#   NVML_TEMPERATURE_THRESHOLD_SHUTDOWN 
#   NVML_TEMPERATURE_THRESHOLD_SLOWDOWN 
#   NVML_TEMPERATURE_THRESHOLD_MEM_MAX 
#   NVML_TEMPERATURE_THRESHOLD_GPU_MAX
#   NVML_TEMPERATURE_THRESHOLD_ACOUSTIC_MIN 
#   NVML_TEMPERATURE_THRESHOLD_ACOUSTIC_CURR <- only this one will be supported
#   NVML_TEMPERATURE_THRESHOLD_ACOUSTIC_MAX 
def set_temperature_thresholds(gpu_handle, threshold_type, temperature_C, dry_run):

    if dry_run != True:
        pynvml.nvmlDeviceSetTemperatureThreshold(gpu_handle, threshold_type, temperature_C)

def print_thresholds_info(configuration):

    gpu_handle = get_GPU_handle(configuration.gpu_name, configuration.gpu_uuid)

    temperarure_thresholds = get_temperarure_thresholds(gpu_handle)

    print(f'{output_separator}')
    #print(f'Temperature threshold - shutdown: {temperarure_thresholds.shutdown}°C')
    #print(f'Temperature threshold - slowdown: {temperarure_thresholds.slowdown}°C')
    #print(f'Temperature threshold - max memory temperature: {temperarure_thresholds.max_memory}°C')
    #print(f'Temperature threshold - ignore base clock: {temperarure_thresholds.gpu_max}°C')
    print(f'Temperature threshold - current acoustic: {temperarure_thresholds.current_acoustic}°C')
    print(f'Temperature threshold - minimum acoustic: {temperarure_thresholds.min_acoustic}°C')
    print(f'Temperature threshold - maximum acoustic: {temperarure_thresholds.max_acoustic}°C')
    print(f'{output_separator}')

def temp_control_subroutine(gpu_handle, target_acoustic_temp_limit, dry_run, verbose):

    current_temp_thresholds = get_temperarure_thresholds(gpu_handle)

    setting_changed = True
    log_msg = []

    log_msg.append(f'Current acoustic threshold: {current_temp_thresholds.current_acoustic}°C')

    if target_acoustic_temp_limit < current_temp_thresholds.min_acoustic or target_acoustic_temp_limit > current_temp_thresholds.max_acoustic:
        log_msg.append(f'WARNING: trying to set acoustic threshold outside of the min({current_temp_thresholds.min_acoustic}°C) and max({current_temp_thresholds.max_acoustic}°C) range')

    if target_acoustic_temp_limit != current_temp_thresholds.current_acoustic:
        set_temperature_thresholds(gpu_handle, pynvml.NVML_TEMPERATURE_THRESHOLD_ACOUSTIC_CURR, target_acoustic_temp_limit, dry_run)
        log_msg.append(f'Setting acoustic temperature threshold: {target_acoustic_temp_limit}°C')
        setting_changed = True
    
    else:
        log_msg.append(f'Nothing to do, current temperature threshold is the same as the target')
        setting_changed = False

    # Only print log messages when necessary to avoid taking too much disk space
    if verbose == True or setting_changed == True:
        log_helper(f'{"\n" + "\n".join(log_msg) + "\n"}')


def control_all(configuration):

    gpu_handle = get_GPU_handle(configuration.gpu_name, configuration.gpu_uuid)
    print_GPU_info(gpu_handle)

    while(True):

        # The user has enabled it with an option to take effect
        if configuration.power_limit != 0:
            power_control_subroutine(gpu_handle, configuration.power_limit, configuration.dry_run, configuration.verbose)

        # The user has enabled it with an option to take effect
        if configuration.acoustic_temp_limit != 0:
            temp_control_subroutine(gpu_handle, configuration.acoustic_temp_limit, configuration.dry_run, configuration.verbose)

        # The user has enabled it with an option to take effect
        if len(configuration.temp_speed_pair) != 0:
            fan_control_subroutine(gpu_handle, configuration)

        if configuration.single_use == True:
            break

        time.sleep(configuration.time_interval)